local AceTimer = LibStub("AceTimer-3.0")
local AceComm = LibStub("AceComm-3.0")
local Serializer = LibStub("AceSerializer-3.0")
local Compressor = LibStub("LibCompress")
local Encoder = Compressor:GetAddonEncodeTable()

-- Prefixes
local STATE_MSG_PREFIX = "BL_STATE_MSG"
local REQUEST_MSG_PREFIX = "BL_REQUEST_MSG"
local UPDATE_MSG_PREFIX = "BL_UPDATE_MSG"

local attemptCounter
local playerName = GetUnitName("player", false)
local updaterTimer
local gottenUpdateMessage = false
local sessionBlacklist = {}

local syncTime
local sessionLastSync = {}

local frame = CreateFrame("Frame")

local function StripRealm(name)
  return gsub(name, "-.*", "") 
end

local function SerializeTable(table)
  local data = Serializer:Serialize(table)
  
  local hash = Compressor:fcs32init()
  hash = Compressor:fcs32update(hash, data)
  hash = Compressor:fcs32final(hash)
  hash = string.format("%010s", tostring(hash)) -- Pad the hash
  data = data .. hash
  
  local compData = Compressor:CompressHuffman(data)
  local encData = Encoder:Encode(compData)
  
  return encData
end

local function DeserializeTable(text)
  local compData = Encoder:Decode(text)
  
  -- Decompress
  local data, msg = Compressor:Decompress(compData)
  if data == nil then
    --print("BossList: Failed to decompress - " .. msg)
    --print("BossList: Please tell Saberos")
    return false;
  end
  
  -- Extract hash
  local msgHash = strsub(data, strlen(data) - 9)
  msgHash = tonumber(msgHash)
  
  if msgHash ~= nil then -- Nil when the number conversion fails, either due to corruption or it being missing
    local hash = Compressor:fcs32init()
    hash = Compressor:fcs32update(hash, strsub(data, 1, strlen(data) - 10))
    hash = Compressor:fcs32final(hash)
    
    if hash ~= msgHash then
      --print("Hash mismatch!")
      return false
    end
  end  
  
  local succes, table = Serializer:Deserialize(data)  
  if not succes then
    --print("BossList: Failed to deserialize - " .. table)
    --print("BossList: Please tell Saberos")
    return false
  end
  
  return true, table
end

-- Send table using prefix
local function SendTable(prefix, table, target) 
  table.revision = BossList.REVISION
  
  local data = SerializeTable(table)
  
  -- If a target is provided, whisper the message. Otherwise broadcast to guild.
  if target == nil then
    AceComm:SendCommMessage(prefix, data, "GUILD")
  else
    AceComm:SendCommMessage(prefix, data, "WHISPER", target)
  end
  
  table.revision = nil
end

local function UpdateHandler(prefix, text, distribution, sender)
  sender = StripRealm(sender)
  if sender == playerName or tContains(sessionBlacklist, sender) then
    return
  end
  
  --print("Received update - " .. prefix .. " from " .. sender)
  local succes, table = DeserializeTable(text)
  
  if not(succes) then
    return AceComm:SendCommMessage(REQUEST_MSG_PREFIX, BossList.lastUpdates[sender], "WHISPER", sender)
  end
  
  if table.revision == nil or table.revision < BossList.REVISION then
    tinsert(sessionBlacklist, sender)
    print("BossList: Received message from older version from " .. sender .. ". Please tell him/her to update.")
    return
  elseif table.revision > BossList.REVISION then
    if not(gottenUpdateMessage) then
      print("BossList: Received message from newer version. Please update your BossList")
      gottenUpdateMessage = true
    end
    return
  else
    table.revision = nil
  end
  
  -- Update local data
  BossList.MergeData(table, sender)
end

-- Check if the sender has made recent changes, and see if his data from you is outdated.
local function MessageHandler(prefix, text, distribution, sender)  
  sender = StripRealm(sender)
  -- Ignore messages from the player himself, or players found to running older versions
  if sender == playerName or tContains(sessionBlacklist, sender) then
    return
  end
  
  local succes, table = DeserializeTable(text)
  if not(succes) then
    sessionLastSync[sender] = nil
    return
  end
  
  -- Check timestamp
  if sessionLastSync[sender] == table.timestamp then
    return
  end
  sessionLastSync[sender] = table.timestamp
  table.timestamp = nil
  
  -- Sender has outdated data of this user, send changes
  if (table[playerName] or 0) < BossList.lastUpdates[playerName] then
    local changesTable = BossList.ChangesSince(table[playerName] or 0)
    SendTable(UPDATE_MSG_PREFIX, changesTable, sender)
  end
  
  -- Sender has newer data, request update
  if (table[sender] or 0) > (BossList.lastUpdates[sender] or 0) then
    AceComm:SendCommMessage(REQUEST_MSG_PREFIX, tostring(table[sender]), "WHISPER", sender)
  end
end

local function RequestHandler(prefix, text, distribution, sender)
  sender = StripRealm(sender)
  
  if sender == playerName or tContains(sessionBlacklist, sender) then
    return
  end
  
  -- Send changes since time in message
  local time = tonumber(text)
  
  local changesTable = BossList.ChangesSince(time)
  SendTable(UPDATE_MSG_PREFIX, changesTable, sender)
end

local function TransmitState()
  BossList.lastUpdates.timestamp = syncTime
  SendTable(STATE_MSG_PREFIX, BossList.lastUpdates) 
  BossList.lastUpdates.timestamp = nil
end

-- Attempt to synchronize 5 times. Multiple attempts are required as on logon chat and guild function calls return
-- false information or simply fail.
local function synchronize()
  if IsInGuild() and GetNumGuildMembers() > 0 then
    
    local _, numOnlineMembers = GetNumGuildMembers()
    if numOnlineMembers > 0  then
      TransmitState()
    else  -- If nobody is online, fail immediately
      AceTimer:CancelTimer(updaterTimer)
    end
    
  end
  
  attemptCounter = attemptCounter + 1
  if attemptCounter >= 5 then
    AceTimer:CancelTimer(updaterTimer)
  end
end

function BossList.Synchronize()
  attemptCounter = 0
  syncTime = time()
  updaterTimer = AceTimer:ScheduleRepeatingTimer(synchronize, 5)
end

function BossList.SendUpdate(table)
  SendTable(UPDATE_MSG_PREFIX, table)
end

AceComm:RegisterComm(STATE_MSG_PREFIX, MessageHandler)
AceComm:RegisterComm(REQUEST_MSG_PREFIX, RequestHandler)
AceComm:RegisterComm(UPDATE_MSG_PREFIX, UpdateHandler)